@@ -1,23 +1,24 @@
 # -*- coding: utf-8 -*-
 
-'''
+"""
 Das Modul definiert alle Exceptions des Frameworks
-'''
+"""
 from __future__ import absolute_import
 
 import sys
-from . import alfredlog
+import logging
 import traceback
 
 from . import settings
 
 
-logger = alfredlog.getLogger(__name__)
+logger = logging.getLogger(__name__)
+
+if settings.experiment.type == "qt":
 
-if settings.experiment.type == 'qt':
     def excepthook(type, value, tb):
-        s = 'Unhandled exception: %s (%s)\n' % (type, value)
-        s = s + 'Traceback:\n' + ''.join(traceback.format_tb(tb))
+        s = "Unhandled exception: %s (%s)\n" % (type, value)
+        s = s + "Traceback:\n" + "".join(traceback.format_tb(tb))
         # logging.critical(s)
         logger.critical(s)
 
@@ -25,9 +26,9 @@ if settings.experiment.type == 'qt':
 
 
 class AlfredError(Exception):
-    u'''
+    u"""
     Jede Exception des Frameworks ist von dieser Klasse abgeleitet.
-    '''
+    """
     pass
 
 
